#!/usr/bin/python3
# -*- coding: utf-8 -*-
import os
import cffi
import numpy as np
import matplotlib.pyplot as plt
from scipy.interpolate import CubicSpline
from math import exp

def polynom (coeff, x):
  y = 0.0
  for c in coeff: y *= x; y += c;
  return y
################################ FORWARD #########################################################
fcf_l = [-1.6322697486E-20,-1.9889266878E-17,-1.0451609365E-14,-3.1088872894E-12,-5.7410327428E-10,
         -6.7509059173E-08,-4.9904828777E-06,-3.2858906784E-04,+2.3622373598E-02,+3.9450128025E+01, 0.0E00]

fcf_h = [-1.2104721275E-23,+9.7151147152E-20,-3.2020720003E-16,+5.6075059059E-13,-5.6072844889E-10,
         +3.1840945719E-07,-9.9457592874E-05,+1.8558770032E-02,+3.8921204975E+01,-1.7600413686E+01]

def exp_part (x):
  c0 = +1.185976E+02
  c1 = -1.183432E-04
  y  = c0 * exp (c1 * (x - 126.9686)**2)
  return y;
def TCK_ctouv (x):
  if (x < 0.0):       return polynom (fcf_l, x)
  else:               return polynom (fcf_h, x) + exp_part (x)
################################ INVERSE #########################################################
'''
icf_l = [-5.1920577E-28,-1.0450598E-23,-8.6632643E-20,-3.7342377E-16,-8.9773540E-13,-1.0833638E-06,
         -1.1662878E-06, 2.5173462E-02, 0.0] # původní, někde je asi chyba
'''
icf_l = [-2.92671625e-28,-4.70434523e-24,-2.62981016e-20,-3.48006363e-17,+1.96290765e-13,+9.54931267e-10,
         +9.03222890e-07,+2.61661840e-02,+1.56311084e-01]

icf_m = [-1.052755E-35, +1.057734E-30, -4.413030E-26, +9.804036E-22, -1.228034E-17, +8.315270E-14,
         -2.503131E-10, +7.860106E-08, +2.508355E-02, +0.0000000E+00]
icf_h = [-3.110810E-26, +8.802193E-21, -9.650715E-16, +5.464731E-11, -1.646031E-06, +4.830222E-02,
         -1.318058E+02]
def TCK_uvtoc (x):
  if   (x < 0.0):     return polynom (icf_l, x)
  elif (x < 20644.0): return polynom (icf_m, x)
  else:               return polynom (icf_h, x)
##################################################################################################
header = '''extern "C" {
#include "tck.h"
};
typedef float real;
/*! @class SplineSet
    @brief Výsledné koeficienty pro výpočet iterace pomocí polynomu 3. řádu.
*/
struct SplineSet {
  real x;   //!< začátky intervalů x
  real a;   //!< záčátky intervalů y, koeficienty a
  real b;   //!< koeficienty b (x^1)
  real c;   //!< koeficienty c (x^2)
  real d;   //!< koeficienty d (x^3)
};
//! V podstatě statický vektor pro SplineSet (bez nutnosti dymanické alokace)
class SPLINE {
  const SplineSet * data;   // zde jsou schovány koeficienty (privátně)
  const int         N;
  public:
    //! Konstruktor
    explicit SPLINE (const SplineSet * p, const int n) noexcept : data(p), N(n) {}
    const real interpolate (const real x) const {
      const unsigned    n = range (x);                     // určíme interval ve kterém budeme počítat
      const SplineSet & s = data [n];                      // koeficienty polynomu v tomto intervalu
      const real        r = x - s.x;                       // souřadnice x v tomto intervalu
      return cubic_poly (s, r);                            // vlastní výpočet
    }
  protected:
    const unsigned range (const real x) const {
      int l=0, r=N - 1u;
      while (l <= r) {
        const int s = (l + r) >> 1;
        const real x0 = data[s + 0].x;
        const real x1 = data[s + 1].x;
        if      (x <  x0) r = s - 1;
        else if (x >= x1) l = s + 1;
        else return s;
      }
      return N - 1u;
    }
    const real cubic_poly (const SplineSet & s, const real x) const {
      return ((s.d * x + s.c) * x + s.b) * x + s.a;
    }
};
'''
c_code = '''
static const SPLINE Thermocouple_forward (C_to_uV, C_to_uV_len);
static const SPLINE Thermocouple_inverse (uV_to_C, uV_to_C_len);

double ComputeForward (double x) {
  const double result = Thermocouple_forward.interpolate ((float) x);
  return result;
}
double ComputeInverse (double x) {
  const double result = Thermocouple_inverse.interpolate ((float) x);
  return result;
}
'''
##################################################################################################
def compute_low (): # vyhodí lepší koeficienty icf_l, ty původní jsou asi nepoužitelné
  xs = np.arange(-200.0, 1.0, 1.0)
  ys = []
  for x in xs: ys.append(TCK_ctouv (x))
  approx = np.polyfit (ys, xs, 8)
  print (approx)
  return np.poly1d(approx)
def cubic_splines (ox, cs, name):
  mx = len (ox) - 1
  ss = []
  for n in range (mx): ss.append ([])
  for n in range (mx):
    ss[n].append(ox[n])
    for k in range (0,4):
      ss[n].append(cs.c.item(3-k,n))
  s = 'static const SplineSet {:s}[] = {{\n'.format(name)
  for p in ss:
    s += '  {'
    for e in p:
      s += ' {:+15.8e}f,'.format(e)
    s += ' },\n'
  s += '};\n'
  s += 'static const int {:s}_len = {:d};\n'.format(name,mx)
  return s
def forward_test(C):
  xs = np.arange(-200.0, 1250.0, 1.0)
  err = []
  for x in xs: err.append (TCK_ctouv (x) - C.ComputeForward(x))
  fig,ap = plt.subplots (1, figsize=(6.0,4.0))
  fig.suptitle ('K-type forward', fontsize=15)
  ap.plot (xs, err)
  ap.legend (['spline error'])
  plt.grid()
  plt.xlabel('temperature [°C]')
  plt.ylabel('voltage [µV]')
  plt.savefig('forward.png')
  #plt.show()
def inverse_test(C):
  xs = np.arange(-5000.0, 50000.0, 100.0)
  err = []
  for x in xs: err.append (TCK_uvtoc (x) - C.ComputeInverse(x))
  fig,ap = plt.subplots (1, figsize=(6.0,4.0))
  fig.suptitle ('K-type inverse', fontsize=15)
  ap.plot (xs, err)
  ap.legend (['spline error'])
  plt.grid()
  plt.ylabel('temperature [°C]')
  plt.xlabel('voltage [µV]')
  plt.savefig('inverse.png')
  #plt.show()

def compute_splines ():
  xs = [-200.0, -175.0, -150.0, -100.0, -50.0, 0.0, 50.0, 100.0, 150.0, 200.0, 250.0, 300.0,
        350.0, 400.0, 500.0, 600.0, 700.0, 800.0, 900.0, 1000.0, 1100.0, 1200.0, 1300.0]
  ys = []
  ss = header
  for x in xs: ys.append (TCK_ctouv (x))
  cs  = CubicSpline(xs, ys)
  ss += cubic_splines (xs, cs, 'C_to_uV')
  cs  = CubicSpline(ys, xs)
  ss += cubic_splines (ys, cs, 'uV_to_C')
  ss += c_code
  out = open ('spline.cpp','w')
  out.write(ss)
  out.close()
  
def spline_tests():
  if os.system('make all') != 0: return
  ffi = cffi.FFI()
  ffi.cdef(''.join(open("tck.h").readlines()))
  C = ffi.dlopen("./spline.so")
  forward_test(C)
  inverse_test(C)
############################  MAIN  ##############################################################
def main_func():
  #rev = compute_low()
  compute_splines()
  spline_tests()
  return
if __name__ == '__main__':
  main_func()
  print ("END OK")
